from hashlib import sha3_256
import colorsys


templates = {}

templates['Section_Begin'] = '<uu5string/><UU5.Bricks.Lsi><UU5.Bricks.Lsi.Item language="%language_code%">'
templates['TSDefinition_Begin'] = '<div id="%TSDefinitionCode%"><UU5.Bricks.Link href="#%TSDefinitionCode%" target="_self"><UU5.Bricks.Icon icon="mdi-blank"/></UU5.Bricks.Link></div>'
templates['TSDefinition_Block'] = '<Uu5Bricks.Block baseUri="https://www.plus4u.net" headerSeparator footerSeparator id="0" info="TSDefinition" ' \
'style="borderColor: %border_color%;" footer="<uu5string/><span style=~"<uu5json/>{~~~"color~~~": ~~~"rgb(158, 158, 158)~~~"}~">%footer%</span>" ' \
'margin=0 contentPadding="0 b" header="%TSDefinitionCode%" collapsible>'
templates['TSDefinition_BlockSectionStart'] = '<Uu5Bricks.Section headerSeparator=false contentPadding="0 b">'
templates['TSDefinition_BlockContent'] = '<UU5.RichText.Block uu5string="<uu5string/>' \
'<strong>Name:</strong> %name%' \
'<UU5.Bricks.Div><strong>Short Name:</strong> %short_name%</UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>Description:</strong> %description%</UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>Data Provider:</strong> %dataprovider%</UU5.Bricks.Div>'
templates['TSDefinition_BlockContentFormulaDifferentialCalculationYes'] = '<UU5.Bricks.Div><strong>Formula computed using differential calculation:</strong> ' \
'<UU5.Bricks.Span style=~"<uu5json/>{~~~"backgroundColor~~~": ~~~"#C8E6C9~~~"}~">Yes</UU5.Bricks.Span></UU5.Bricks.Div>'
templates['TSDefinition_BlockContentFormulaDifferentialCalculationNo'] = '<UU5.Bricks.Div><strong>Formula computed using differential calculation:</strong> ' \
'<UU5.Bricks.Span>No</UU5.Bricks.Span></UU5.Bricks.Div>'
templates['TSDefinition_BlockSectionEnd'] = '"/></Uu5Bricks.Section><Uu5Bricks.Section headerSeparator=false margin=0 contentPadding=0>'
templates['TSDefinition_Formula'] = '<Uu5CodeKitBricks.Code value="%formula%" codeStyle="javascript" maxRows=50 indent=4/>'
templates['TSDefinition_PredecessorsStart'] = '<Uu5Bricks.Layout type="columns"><Uu5Bricks.Layout.Item colSpan="m: 6">' \
'<UuContentKit.Bricks.BlockDefault icon="mdi-arrow-collapse-right">' \
'<UU5.RichText.Block uu5string="<uu5string/><UU5.Bricks.Div><UU5.Bricks.Span style=\\\"<uu5json/>{\\\\\\\"color\\\\\\\": \\\\\\\"#B71C1C\\\\\\\"' \
'}\\\"><strong>Predecessors (total %predecessors_count%):</strong></UU5.Bricks.Span></UU5.Bricks.Div><UU5.Bricks.Ul>'
templates['TSDefinition_Predecessor'] = '<UU5.Bricks.Li>' \
'<UU5.Bricks.Link href=~"%predecessor_link%~" target=~"_self~">%predecessor%</UU5.Bricks.Link></UU5.Bricks.Li>'
templates['TSDefinition_PredecessorsEnd'] = '</UU5.Bricks.Ul>"/></UuContentKit.Bricks.BlockDefault></Uu5Bricks.Layout.Item>'
templates['TSDefinition_SuccessorsStart'] = '<Uu5Bricks.Layout.Item colSpan="m: 6">' \
'<UuContentKit.Bricks.BlockDefault icon="mdi-arrow-expand-right">' \
'<UU5.RichText.Block uu5string="<uu5string/><UU5.Bricks.Div><UU5.Bricks.Span style=\\\"<uu5json/>{\\\\\\\"color\\\\\\\": \\\\\\\"#1B5E20\\\\\\\"' \
'}\\\"><strong>Successors (total %SuccessorsCount%):</strong></UU5.Bricks.Span></UU5.Bricks.Div><UU5.Bricks.Ul>'
templates['TSDefinition_Successor'] = '<UU5.Bricks.Li>' \
'<UU5.Bricks.Link href=~"%successor_link%~" target=~"_self~">%successor%</UU5.Bricks.Link></UU5.Bricks.Li>'
templates['TSDefinition_SuccessorsEnd'] = '</UU5.Bricks.Ul>"/></UuContentKit.Bricks.BlockDefault></Uu5Bricks.Layout.Item></Uu5Bricks.Layout>'
templates['TSDefinition_End'] = '</Uu5Bricks.Section></Uu5Bricks.Block>'
templates['Section_End'] = '</UU5.Bricks.Lsi.Item></UU5.Bricks.Lsi>'

templates['TSObjectType_Separator'] = '<br/><br/><UU5.Bricks.Line colorSchema="grey-rich" size="m"/>'
templates['TSObjectType_Begin'] = '<div id="%TSObjectTypeCode%"><UU5.Bricks.Link href="#%TSObjectTypeCode%" target="_self"><UU5.Bricks.Icon icon="mdi-blank"/></UU5.Bricks.Link></div>'
templates['TSObjectType_Block'] = '<Uu5Bricks.Block baseUri="https://www.plus4u.net" headerSeparator footerSeparator id="0" info="TSObjectType" ' \
'style="borderColor: %border_color%;" footer="<uu5string/><span style=~"<uu5json/>{~~~"color~~~": ~~~"rgb(158, 158, 158)~~~"}~">%footer%</span>" ' \
'margin=0 contentPadding="0 b" header="%TSObjectTypeCode%" collapsible significance="highlighted" colorScheme="%color_scheme%">'
templates['TSObjectType_BlockSectionStart'] = '<Uu5Bricks.Section headerSeparator=false contentPadding="0 b">'
templates['TSObjectType_BlockContent'] = '<UU5.RichText.Block uu5string="<uu5string/>' \
'<strong>Name:</strong> %name%' \
'<UU5.Bricks.Div><strong>Short Name:</strong> %short_name%</UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>Description:</strong> %description%</UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>TSDefinitions count:</strong> %tsdefinitions_count%</UU5.Bricks.Div>'
templates['TSObjectType_DimensionsBegin'] = '<UU5.Bricks.Div><strong>Dimensions (total %dimensions_count%):</strong></UU5.Bricks.Div><UU5.Bricks.Ul>'
templates['TSObjectType_Dimension'] = '<UU5.Bricks.Li>' \
'<UU5.Bricks.Link href=~"%dimension_link%~" target=~"_self~">%dimension%</UU5.Bricks.Link></UU5.Bricks.Li>'
templates['TSObjectType_DimensionsEnd'] = '</UU5.Bricks.Ul>'
templates['TSObjectType_BlockContentEnd'] = '"/>'
templates['TSObjectType_TransformationsToSectionStart'] = '<Uu5Bricks.Layout type="columns"><Uu5Bricks.Layout.Item colSpan="m: 6">' \
'<UuContentKit.Bricks.BlockDefault icon="mdi-arrow-collapse-right">' \
'<UU5.RichText.Block uu5string="<uu5string/><UU5.Bricks.Div><UU5.Bricks.Span style=\\\"<uu5json/>{\\\\\\\"color\\\\\\\": \\\\\\\"#B71C1C\\\\\\\"' \
'}\\\"><strong>Transformations to this TSObjectType (total %predecessors_count%):</strong></UU5.Bricks.Span></UU5.Bricks.Div><UU5.Bricks.Ul>'
templates['TSObjectType_TransformationsToStart'] = '<UU5.Bricks.Li>' \
'<UU5.Bricks.Link href=~"%predecessor_link%~" target=~"_self~">%transformation_code%</UU5.Bricks.Link></UU5.Bricks.Li>'
templates['TSObjectType_TransformationsToEnd'] = '</UU5.Bricks.Ul>"/></UuContentKit.Bricks.BlockDefault></Uu5Bricks.Layout.Item>'
templates['TSObjectType_TransformationsFromSectionStart'] = '<Uu5Bricks.Layout.Item colSpan="m: 6">' \
'<UuContentKit.Bricks.BlockDefault icon="mdi-arrow-expand-right">' \
'<UU5.RichText.Block uu5string="<uu5string/><UU5.Bricks.Div><UU5.Bricks.Span style=\\\"<uu5json/>{\\\\\\\"color\\\\\\\": \\\\\\\"#1B5E20\\\\\\\"' \
'}\\\"><strong>Transformations from this TSObjectType (total %successors_count%):</strong></UU5.Bricks.Span></UU5.Bricks.Div><UU5.Bricks.Ul>'
templates['TSObjectType_TransformationsFromStart'] = '<UU5.Bricks.Li>' \
'<UU5.Bricks.Link href=~"%successor_link%~" target=~"_self~">%transformation_code%</UU5.Bricks.Link></UU5.Bricks.Li>'
templates['TSObjectType_TransformationsFromEnd'] = '</UU5.Bricks.Ul>"/></UuContentKit.Bricks.BlockDefault></Uu5Bricks.Layout.Item></Uu5Bricks.Layout>'
templates['TSObjectType_End'] = '</Uu5Bricks.Section></Uu5Bricks.Block>'

templates['TSObjectTypeRelation_Separator'] = '<br/><br/><UU5.Bricks.Line colorSchema="grey-rich" size="m"/>'
templates['TSObjectTypeRelation_Begin'] = '<div id="%TSObjectTypeRelationCode%"><UU5.Bricks.Link href="#%TSObjectTypeRelationCode%" target="_self"><UU5.Bricks.Icon icon="mdi-blank"/></UU5.Bricks.Link></div>'
templates['TSObjectTypeRelation_Block'] = '<Uu5Bricks.Block baseUri="https://www.plus4u.net" headerSeparator footerSeparator id="0" info="TSObjectTypeRelation" ' \
'style="borderColor: %border_color%;" footer="<uu5string/><span style=~"<uu5json/>{~~~"color~~~": ~~~"rgb(158, 158, 158)~~~"}~">%footer%</span>" ' \
'margin=0 contentPadding="0 b" header="%TSObjectTypeRelationCode%" collapsible significance="distinct" colorScheme="%color_scheme%">'
templates['TSObjectTypeRelation_BlockSectionStart'] = '<Uu5Bricks.Section headerSeparator=false contentPadding="0 b">'
templates['TSObjectTypeRelation_BlockContent'] = '<UU5.RichText.Block uu5string="<uu5string/>' \
'<strong>Name:</strong> %name%' \
'<UU5.Bricks.Div><strong>Short Name:</strong> %short_name%</UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>Description:</strong> %description%</UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>Relation Type:</strong> %relation_type%</UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>Source: </strong><UU5.Bricks.Link href=~"%source_tsobjecttype_link%~" target=~"_self~">%source_tsobjecttype%</UU5.Bricks.Link></UU5.Bricks.Div>' \
'<UU5.Bricks.Div><strong>Target: </strong><UU5.Bricks.Link href=~"%target_tsobjecttype_link%~" target=~"_self~">%target_tsobjecttype%</UU5.Bricks.Link></UU5.Bricks.Div>"/>'
templates['TSObjectTypeRelation_End'] = '</Uu5Bricks.Section></Uu5Bricks.Block>'


def escape_str(value: str, use_br_for_newlines: bool) -> str:
    """
    Escape special characters in a string for HTML display.
    """
    result = value
    result = result.replace("\\", "\\\\").replace('\"', '\\\"').replace('\'', '\\\'')
    if use_br_for_newlines:
        result = result.replace("\r", "").replace("\n", "<br/>")
    else:
        result = result.replace("\r", "")
    return result

def use_template(template_name: str, replacements: dict[str, str], language: str, use_br_for_newlines: bool = True) -> str:
    """
    Replace placeholders in the template with actual values from replacements dictionary.
    Placeholders are in the format %placeholder%.
    """
    result = templates[template_name]
    result = result.replace("~", "\\")
    for key, value in replacements.items():
        placeholder = f"%{key}%"
        if isinstance(value, str):
            new_value = escape_str(value, use_br_for_newlines)           
            result = result.replace(placeholder, new_value)
        elif isinstance(value, dict) and language in value:
            new_value = escape_str(value[language], use_br_for_newlines)
            result = result.replace(placeholder, new_value)
        elif value is None:
            result = result.replace(placeholder, "--")
        else:
            raise ValueError(f"Unsupported replacement type for key '{key}': {type(value)}")
    return result

class Palette:
    @staticmethod
    def get_palette() -> list[str]:
        """
        Return a predefined palette of colors.
        """
        return [
            "#008000", "#000080", "#808000", "#008080", "#800080",
            "#FF0000", "#00FF00", "#0000FF", "#FF5733", "#303030",
            "#FFFF00", "#00FFFF", "#FF00FF", "#808080", "#800000",
            "#E27D60", "#85DCB0", "#E8A87C", "#C38D9E", "#41B3A3",
            "#6B5B95", "#FF6F61", "#88B04B", "#F7CAC9", "#92A8D1"
        ]
    
    @staticmethod
    def get_color(value: str) -> str:
        """
        Get a consistent color from the palette based on the hash of the input value.
        """
        value_hash = sha3_256(value.encode('utf-8')).hexdigest()
        # palette = Palette.get_palette()
        # index = int(value_hash, 16) % len(palette)
        # return palette[index]
        index = int(value_hash, 16) % 1080
        hue = index / 1080.0
        lightness = 0.3 + (index % 3) * 0.2
        rgb = colorsys.hls_to_rgb(hue, lightness, 1.0)
        return f"#{int(rgb[0]*255):02X}{int(rgb[1]*255):02X}{int(rgb[2]*255):02X}"



from pathlib import Path
from .doctsdefinition import DocTsDefinition
from .doctsformula import DocTsFormula
from .doctsobjecttype import DocTsObjectType, DocDimension
from .doctsobjecttyperelation import DocTsObjectTypeRelation
from .doctstype import DocTsType
from .doctsbusinesstype import DocTsBusinessType
from .doctstimeunit import DocTsTimeUnit
from .doctemplates import use_template
import fetchx.ioutils as ioutils


class Documentation:
    def __init__(self, metamodel_zip_path: Path, documentation_path: Path):
        """
        Initialize the Documentation object by loading the metamodel from the specified zip file.
        :param metamodel_zip_path: Path to the metamodel zip file.
        :param documentation_path: Path where the documentation will be generated.
        """
        self.metamodel_path = documentation_path / "metamodel"
        ioutils.delete_folder(str(self.metamodel_path.resolve()))
        # if metamodel_zip_path is a directory, use it directly
        if metamodel_zip_path.is_dir():
            self.metamodel_path = metamodel_zip_path
        else:
            self._unzip_to_directory(metamodel_zip_path, self.metamodel_path)
        self.documentation_path = documentation_path
        self.tsdefinitions = DocTsDefinition.load_from_directory(self.metamodel_path)
        self.tsformulas = DocTsFormula.load_from_directory(self.metamodel_path)
        self.tsobjecttypes = DocTsObjectType.load_from_directory(self.metamodel_path)
        self.tsobjecttypesrelations = DocTsObjectTypeRelation.load_from_directory(self.metamodel_path)
        self.tstypes = DocTsType.load_from_directory(self.metamodel_path)
        self.tsbusinesstypes = DocTsBusinessType.load_from_directory(self.metamodel_path)
        self.tstimeunits = DocTsTimeUnit.load_from_directory(self.metamodel_path)
        self._link_objects()
        self._link_tsdefinition_predecessors_and_successors()

    def _unzip_to_directory(self, zip_path: Path, extract_to: Path):
        """
        Unzip the specified zip file to the given directory.
        :param zip_path: Path to the zip file.
        :param extract_to: Directory where the contents will be extracted.
        """
        import zipfile
        ioutils.create_folder_structure(extract_to)
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(extract_to)

    def _link_objects(self):
        """
        Link the loaded objects based on their references.
        This method should be called after loading all objects.
        """
        # Link to object type relations
        for tsobjecttyperelation_code, tsobjecttyperelation in self.tsobjecttypesrelations.items():
            # Link to source tsobjecttype
            if "sourceTsObjectTypeCode" in tsobjecttyperelation and tsobjecttyperelation["sourceTsObjectTypeCode"] in self.tsobjecttypes:
                source_tsobjecttype = self.tsobjecttypes[tsobjecttyperelation["sourceTsObjectTypeCode"]]
                tsobjecttyperelation.source_tsobjecttype = source_tsobjecttype
                source_tsobjecttype.relationsFromThisTsObjectType.append(tsobjecttyperelation)
            # Link to target tsobjecttype
            if "targetTsObjectTypeCode" in tsobjecttyperelation and tsobjecttyperelation["targetTsObjectTypeCode"] in self.tsobjecttypes:
                target_tsobjecttype = self.tsobjecttypes[tsobjecttyperelation["targetTsObjectTypeCode"]]
                tsobjecttyperelation.target_tsobjecttype = target_tsobjecttype
                target_tsobjecttype.relationsToThisTsObjectType.append(tsobjecttyperelation)

        # Link to dimensions
        for tsobjecttype_code, tsobjecttype in self.tsobjecttypes.items():
            if tsobjecttype["implementationType"] == "tsDimensionList":
                tsobjecttype.dimensions = []
            else:
                tsobjecttype.dimensions = None
                continue
            # link dimensions
            if "tsDimensionList" in tsobjecttype:
                for dimension in tsobjecttype["tsDimensionList"]:
                    if "tsObjectTypeCode" in dimension and dimension["tsObjectTypeCode"] in self.tsobjecttypes:
                        tsobjecttype.dimensions.append(DocDimension(dimension, self.tsobjecttypes[dimension["tsObjectTypeCode"]]))    

        # Link to definitions
        for tsdefinition_code, tsdefinition in self.tsdefinitions.items():
            # Link to parent tsobjecttype
            if "tsObjectTypeCode" in tsdefinition and tsdefinition["tsObjectTypeCode"] in self.tsobjecttypes:
                parent_tsobjecttype = self.tsobjecttypes[tsdefinition["tsObjectTypeCode"]]
                tsdefinition.parent_tsobjecttype = parent_tsobjecttype
                parent_tsobjecttype.tsdefinitions.append(tsdefinition)
            # Link to tstype
            if "tsTypeCode" in tsdefinition and tsdefinition["tsTypeCode"] in self.tstypes:
                tstype = self.tstypes[tsdefinition["tsTypeCode"]]
                tsdefinition.tstype = tstype
            # Link to tstimeunit
            if "tstimeunitCode" in tsdefinition and tsdefinition["tstimeunitCode"] in self.tstimeunits:
                tstimeunit = self.tstimeunits[tsdefinition["tstimeunitCode"]]
                tsdefinition.tstimeunit = tstimeunit

        for tstype_code, tstype in self.tstypes.items():
            # Link to business type
            if "tsBusinessTypeCode" in tstype and tstype["tsBusinessTypeCode"] in self.tsbusinesstypes:
                tsbusinesstype = self.tsbusinesstypes[tstype["tsBusinessTypeCode"]]
                tstype.tsbusinesstype = tsbusinesstype
            # Link to time unit
            if "tsTimeUnitCode" in tstype and tstype["tsTimeUnitCode"] in self.tstimeunits:
                tstimeunit = self.tstimeunits[tstype["tsTimeUnitCode"]]
                tstype.tstimeunit = tstimeunit

        for tsbusinesstype_code, tsbusinesstype in self.tsbusinesstypes.items():
            # Link to time unit
            if "tsTimeUnitCode" in tsbusinesstype and tsbusinesstype["tsTimeUnitCode"] in self.tstimeunits:
                tstimeunit = self.tstimeunits[tsbusinesstype["tsTimeUnitCode"]]
                tsbusinesstype.tstimeunit = tstimeunit
        
        # Link to formulas
        for tsformula_code, tsformula in self.tsformulas.items():
            # Link to tsdefinition
            if "tsDefinitionCode" in tsformula and tsformula["tsDefinitionCode"] in self.tsdefinitions:
                tsdefinition = self.tsdefinitions[tsformula["tsDefinitionCode"]]
                tsformula.tsdefinition = tsdefinition
                tsdefinition.formula = tsformula

    def _link_tsdefinition_predecessors_and_successors(self):
        """
        Link tsdefinitions based on formulas.
        If a formula of a tsdefinition references another tsdefinition code, it is considered a predecessor
        of the current tsdefinition, and the current tsdefinition is a successor of the referenced tsdefinition.
        """
        for tsdefinition_code, tsdefinition in self.tsdefinitions.items():
            if tsdefinition.formula is not None:
                formula_words = tsdefinition.formula.parse_formula_words()
                for word in formula_words:
                    if word in self.tsdefinitions:
                        predecessor = self.tsdefinitions[word]
                        tsdefinition.predecessors.append(predecessor)
                        predecessor.successors.append(tsdefinition)

    def modify_links(self, *, pages: list[str], page_urls: list[str]):
        """
        Modify internal links in the documentation to point to the specified page URLs.
        :param page_urls: List of page URLs to replace internal links with.
        """
        result = []
        # dictionary of links. Example 
        # {"code_anchor": {"old_link", "href='#code_anchor'", "new_link": "href='full_page_url#code_anchor'"}}
        links = {}
        if len(page_urls) < len(pages):
            print("Warning: Not enough page URLs provided to modify links. See --pages parameter")
            return pages
        # find all anchors in the page and map them to full URLs
        for page_index, page in enumerate(pages):
            # find anchors for all tsdefinitions
            for tsdefinition_code in self.tsdefinitions.keys():
                anchor_to_search = f'{tsdefinition_code}'
                old_link = f'href=\"#{anchor_to_search}\"'
                new_link = f'href=\"{page_urls[page_index]}#{anchor_to_search}\"'
                if anchor_to_search in page:
                    links[anchor_to_search] = {
                        'old_link': old_link,
                        'new_link': new_link
                    }
            # find anchors for all tsobjecttypes and tsobjecttyperelations
            for tsobjecttype_code in self.tsobjecttypes.keys():
                anchor_to_search = f'<div id=\"{tsobjecttype_code}\">'
                old_link = f'href=\"#{anchor_to_search}\"'
                new_link = f'href=\"{page_urls[page_index]}#{anchor_to_search}\"'
                if anchor_to_search in page:
                    links[anchor_to_search] = {
                        'old_link': old_link,
                        'new_link': new_link
                    }
            # find anchors for all tsobjecttyperelations
            for tsobjecttyperelation_code in self.tsobjecttypesrelations.keys():
                anchor_to_search = f'<div id=\"{tsobjecttyperelation_code}\">'
                old_link = f'href=\"#{anchor_to_search}\"'
                new_link = f'href=\"{page_urls[page_index]}#{anchor_to_search}\"'
                if anchor_to_search in page:
                    links[anchor_to_search] = {
                        'old_link': old_link,
                        'new_link': new_link
                    }
        # replace anchors in pages with full URLs
        for page in pages:
            modified_page = page
            for code, full_url in links.items():
                modified_page = modified_page.replace(f'href="{code}"', f'href="{full_url}"')
            result.append(modified_page)
        return result

    def generate_documentation(self, *, language: str = "en", page_urls: list[str] = []):
        """
        Generate documentation files for each tsdefinition.
        """
        output_path_tsdefinitions = self.documentation_path / "doc/tsdefinitions"
        output_path_tsobjecttypes = self.documentation_path / "doc/tsobjecttypes"
        output_path_tsobjecttyperelations = self.documentation_path / "doc/tsobjecttyperelations"
        output_path_all = self.documentation_path / "doc"
        ioutils.delete_folder(str(output_path_all.resolve()))
        ioutils.create_folder_structure(output_path_tsdefinitions)
        ioutils.create_folder_structure(output_path_tsobjecttypes)
        ioutils.create_folder_structure(output_path_tsobjecttyperelations)
        ioutils.create_folder_structure(output_path_all)

        mini_sections = []
        # Generate documentation for each tsdefinition
        for tsobjecttype_code, tsobjecttype in self.tsobjecttypes.items():
            mini_sections.append('')
            # generate documentation for each tsobjecttype
            single_tsobjecttype = use_template('Section_Begin',{'language_code':language}, language=language)
            single_tsobjecttype += use_template('TSObjectType_Separator',{'language_code':language}, language=language)
            single_tsobjecttype += tsobjecttype.to_string(language=language)
            single_tsobjecttype += use_template('Section_End',{}, language=language)
            output_file = output_path_tsobjecttypes / f"{tsobjecttype['code']}.txt"
            with open(output_file, 'w', encoding='utf-8') as f:
                f.write(single_tsobjecttype)
            # append to entire page
            mini_sections[-1] += use_template('TSObjectType_Separator',{'language_code':language}, language=language)
            mini_sections[-1] += tsobjecttype.to_string(language=language)
            for tsObjectTypeRelation in tsobjecttype.relationsFromThisTsObjectType:
                single_tsobjecttyperelation = use_template('Section_Begin',{'language_code':language}, language=language)
                single_tsobjecttyperelation += tsObjectTypeRelation.to_string(language=language)
                single_tsobjecttyperelation += use_template('Section_End',{}, language=language)
                mini_sections[-1] += single_tsobjecttyperelation
                output_file = output_path_tsobjecttyperelations / f"{tsObjectTypeRelation['code']}.txt"
                with open(output_file, 'w', encoding='utf-8') as f:
                    f.write(tsObjectTypeRelation.to_string(language=language))
            # generate documentation for each tsdefinition in this tsobjecttype
            for tsdefinition in tsobjecttype.tsdefinitions:
                single_tsdefinition = use_template('Section_Begin',{'language_code':language}, language=language)
                single_tsdefinition += tsdefinition.to_string(language=language)
                single_tsdefinition += use_template('Section_End',{}, language=language)
                mini_sections[-1] += tsdefinition.to_string(language=language)
                output_file = output_path_tsdefinitions / f"{tsdefinition['code']}.txt"
                with open(output_file, 'w', encoding='utf-8') as f:
                    f.write(tsdefinition.to_string(language=language))

        # generate documentation page by page
        pages = ['']
        for mini_section in mini_sections:
            if len(pages[-1]) + len(mini_section) > 485000:
                pages.append('')
            pages[-1] += mini_section
        # modify links to point to full page urls
        pages = self.modify_links(pages=pages, page_urls=page_urls)
        # generate documentation for all tsdefinitions
        for page_index, page in enumerate(pages):
            doc_entirepage = use_template('Section_Begin',{'language_code':language}, language=language)
            doc_entirepage += page
            doc_entirepage += use_template('Section_End',{}, language=language)
            # write to file
            output_file = output_path_all / f"tsdefinitions_{page_index}.txt"
            with open(output_file, 'w', encoding='utf-8') as f:
                f.write(doc_entirepage)
            



from pathlib import Path
import fetchx.ioutils as ioutils
from .doctemplates import Palette, use_template


class DocTsObjectTypeRelation(dict):
    def __init__(self, *args, **kwargs):
        """
        Represents a Time Series Object Type Relation in the metamodel.
        Inherits from dict to store relation properties.
        """
        super().__init__(*args, **kwargs)
        self.source_tsobjecttype = None  # type: DocTsObjectType | None
        self.target_tsobjecttype = None  # type: DocTsObjectType | None
        self._html: str | None = None  # type: str | None

    @staticmethod
    def load_from_directory(base_path: Path) -> dict[str, 'DocTsObjectTypeRelation']:
        """
        Load all Time Series Object Type Relations from the specified directory.
        :param base_path: Base path where the metamodel is located.
        :return: A dictionary mapping relation codes to DocTsObjectTypeRelation instances.
        """
        result = {}
        tsobjecttyperelations_path = base_path / "metamodel/tsObjectTypeRelation"
        # Load all .json files
        files, directories = ioutils.search_folder(str(tsobjecttyperelations_path.resolve()), list_of_extensions=[".json"])
        for file in files:
            value = ioutils.load_json(file)
            code = Path(file).stem
            result[code] = DocTsObjectTypeRelation(value)
        return result    
    
    def to_string(self, language: str = "en") -> str:
        """
        Convert the Time Series Object Type to a string representation.
        :param language: Language code for localization.
        :return: String representation of the Time Series Object Type.
        """
        if self._html is not None:
            return self._html
        result = ""
        # generate tsobjecttype for specific language, e.g., 'en'
        result += use_template('TSObjectTypeRelation_Begin', {'TSObjectTypeRelationCode': self.get("code", "--")}, language=language)
        # Generate footer content
        footer_content = ""
        footer_content += f'<strong>Value Provider Type: </strong>{self.get("valueProviderType", "--")}, '
        footer_content += f'<strong>Value Provider API: </strong>{self.get("valueProviderApi", "--")}'
        # generate block with border color from parent tsobjecttype
        result += use_template('TSObjectTypeRelation_Block',{
            'border_color': self.source_tsobjecttype.color if self.source_tsobjecttype is not None else "grey",
            'footer': footer_content,
            'TSObjectTypeRelationCode':self.get('code','--'),
            'color_scheme': 'yellow' if self.get("valueProviderType", "--") != 'mapping' else 'grey'
            }, language=language
        ) # type: ignore
        result += use_template('TSObjectTypeRelation_BlockSectionStart', {}, language=language)
        # generate block content
        result += use_template('TSObjectTypeRelation_BlockContent',{
            'name': self.get('name', '--'),
            'short_name': self.get('shortName', '--'),
            'description': self.get('desc', '--'),
            'relation_type': self['sourceCardinality'] + " to " + self['targetCardinality'],
            'source_tsobjecttype_link': f'#{self.source_tsobjecttype["code"]}', # type: ignore
            'source_tsobjecttype': self.source_tsobjecttype['code'] if self.source_tsobjecttype is not None else "--",
            'target_tsobjecttype_link': f'#{self.target_tsobjecttype["code"]}', # type: ignore
            'target_tsobjecttype': self.target_tsobjecttype['code'] if self.target_tsobjecttype is not None else "--"
            }, language=language
        )
        result += use_template('TSObjectTypeRelation_End', {}, language=language)
        self._html = result
        return result
        


from pathlib import Path
from .doctsformula import DocTsFormula
from .doctsobjecttype import DocTsObjectType
import fetchx.ioutils as ioutils
from .doctemplates import use_template
from .doctstype import DocTsType
from .doctstimeunit import DocTsTimeUnit 


class DocTsDefinition(dict):
    def __init__(self, *args, **kwargs):
        """
        Represents a Time Series Definition in the metamodel.
        Inherits from dict to store definition properties.
        """
        super().__init__(*args, **kwargs)
        self.parent_tsobjecttype = None # type: DocTsObjectType | None
        self.formula = None # type: DocTsFormula | None
        self.predecessors = [] # type: list[DocTsDefinition]
        self.successors = [] # type: list[DocTsDefinition]
        self.tstype = None # type: DocTsType | None
        self.tstimeunit = None # type: DocTsTimeUnit | None
        self._html = None # type: str | None

    @staticmethod
    def load_from_directory(base_path: Path) -> dict[str, 'DocTsDefinition']:
        """
        Load all Time Series Definitions from the specified directory.
        :param base_path: Base path where the metamodel is located.
        :return: A dictionary mapping definition codes to DocTsDefinition instances.
        """
        result = {}
        tsdefinitions_path = base_path / "metamodel/tsDefinition"
        # Load all .json files
        files, directories = ioutils.search_folder(tsdefinitions_path, list_of_extensions=[".json"])
        for file in files:
            value = ioutils.load_json(file) 
            code = Path(file).stem
            result[code] = DocTsDefinition(value)
        return result
    
    @property
    def time_unit(self) -> str:
        """
        Get the time unit associated with this Time Series Definition.
        :return: Time unit as a string.
        """
        result = "--"
        if self.tstimeunit is not None:
            return str(self.tstimeunit)
        if self.tstype is not None and self.tstype.tstimeunit is not None:
            return str(self.tstype.tstimeunit)
        if self.tstype is not None and self.tstype.tsbusinesstype is not None and \
            self.tstype.tsbusinesstype.tstimeunit is not None:
            return str(self.tstype.tsbusinesstype.tstimeunit)
        return '--'
    
    @property
    def value_unit(self) -> str:
        """
        Get the value unit associated with this Time Series Definition.
        :return: Value unit as a string.
        """
        result = "--"
        if self.tstype is not None and self.tstype.tsbusinesstype is not None:
            value_unit = self.tstype.tsbusinesstype.get('valueUnit', '--')
        return result
    
    @property
    def aggregation_function(self) -> str:
        """
        Get the aggregation function associated with this Time Series Definition.
        :return: Aggregation function as a string.
        """
        if self.tstype is not None and self.tstype.tsbusinesstype is not None and \
            'aggregationFunction' in self.tstype.tsbusinesstype:
            result = self.tstype.tsbusinesstype.get('aggregationFunction', '--')
        elif self.tstype is not None and 'aggregationFunction' in self.tstype:
            result = self.tstype.get('aggregationFunction', '--')
        else:
            result = self.get('aggregationFunction', '--')
        return result
    
    @property
    def disaggregation_function(self) -> str:
        """
        Get the disaggregation function associated with this Time Series Definition.
        :return: Disaggregation function as a string.
        """
        if self.tstype is not None and self.tstype.tsbusinesstype is not None and \
            'disaggregationFunction' in self.tstype.tsbusinesstype:
            result = self.tstype.tsbusinesstype.get('disaggregationFunction', '--')
        elif self.tstype is not None and 'disaggregationFunction' in self.tstype:
            result = self.tstype.get('disaggregationFunction', '--')
        else:
            result = self.get('disaggregationFunction', '--')
        return result
    
    @property
    def data_provider(self) -> str:
        """
        Get the data provider associated with this Time Series Definition.
        :return: Data provider as a string.
        """
        result = str(self.get('dataProvider')) if self.get('dataProvider') is not None else 'manual input or import from an external system'
        return result

    def to_string(self, language: str = "en") -> str:
        """
        Convert the Time Series Definition to a string representation.
        :param language: Language code for localization.
        :return: String representation of the Time Series Definition.
        """
        if self._html is not None:
            return self._html
        # Generate documentation using templates
        result = ''
        code = self.get('code','unknown_unknown_unknown_unknown')
        code_split = code.split('_')
        while len(code_split) < 4:
            code_split.append('unknown')  # pad with unknown
        # Generate footer content
        footer_content = ""
        footer_content += f'<strong>Time Zone: </strong>{self.get("tsTimeZoneCode", "--")}, '
        footer_content += f'<strong>Time Sample: </strong>{self.time_unit}, '
        footer_content += f'<strong>Agg Func: </strong>{self.aggregation_function}, '
        footer_content += f'<strong>Disagg Func: </strong>{self.disaggregation_function}, '
        footer_content += f'<br/>'
        if self.value_unit is not None and self.value_unit != '--':
            footer_content += f'<strong>Unit: </strong>{self.value_unit}, '
        if self.tstype is not None:
            footer_content += f'<strong>TS Type: </strong>{self.tstype.get("code","--")}, '
        if self.tstype is not None and self.tstype.tsbusinesstype is not None:
            footer_content += f'<strong>TS Business Type: </strong>{self.tstype.tsbusinesstype.get("code","--")}, '
        footer_content += f'<strong>Object Type: </strong>{self.get("tsObjectTypeCode", "--")} '
        # generate tsdefinition for specific language, e.g., 'en'
        result += use_template('Section_Begin',{'language_code':language}, language=language)
        result += use_template('TSDefinition_Begin',{'TSDefinitionCode':self.get('code','--')}, language=language)
        # generate block with border color from parent tsobjecttype
        result += use_template('TSDefinition_Block',{
            'border_color': self.parent_tsobjecttype.color,
            'footer': footer_content,
            'TSDefinitionCode':self.get('code','--')
            }, language=language
        ) # type: ignore
        result += use_template('TSDefinition_BlockSectionStart',{}, language=language)
        # fill in block content
        result += use_template('TSDefinition_BlockContent',{
            'name': self.get('name','--'),
            'short_name': self.get('shortName','--'),
            'description': self.get('desc','--'),
            'dataprovider': self.data_provider,
        }, language=language)
        # add formula differential calculation info if applicable
        if self.formula is not None:
            if self.formula.get('differentialCalculation', False):
                result += use_template('TSDefinition_BlockContentFormulaDifferentialCalculationYes', {}, language=language)
            else:
                result += use_template('TSDefinition_BlockContentFormulaDifferentialCalculationNo', {}, language=language)
        result += use_template('TSDefinition_BlockSectionEnd',{}, language=language)
        # add formula section if formula exists
        if self.formula:
            formula_str = str(self.formula.get('formula',''))
            # escape special characters
            #formula_str = formula_str.replace("\\", "\\\\").replace('\"', '\\\"').replace('\'', '\\\'')
            result += use_template('TSDefinition_Formula',{'formula': formula_str}, language=language, use_br_for_newlines=False)
        # if there is at least one predecessor and one successor then display them in two separate sections
        if len(self.predecessors) > 0 or len(self.successors) > 0:
            # add predecessors
            result += use_template('TSDefinition_PredecessorsStart',{
                'predecessors_count': str(len(self.predecessors))
            }, language=language)
            for predecessor in self.predecessors:
                result += use_template('TSDefinition_Predecessor',{
                    'predecessor': predecessor.get('code',''),
                    'predecessor_link': f"#{predecessor.get('code','')}"
                }, language=language)
            result += use_template('TSDefinition_PredecessorsEnd',{}, language=language)
            # add successors
            result += use_template('TSDefinition_SuccessorsStart',{
                'SuccessorsCount': str(len(self.successors))
            }, language=language)
            for successor in self.successors:
                result += use_template('TSDefinition_Successor',{
                    'successor': successor.get('code',''),
                    'successor_link': f"#{successor.get('code','')}"
                }, language=language)
            result += use_template('TSDefinition_SuccessorsEnd',{}, language=language)
        # add end section
        result += use_template('TSDefinition_End',{}, language=language)
        result += use_template('Section_End',{}, language=language)
        self._html = result + "\n"
        return self._html




from pathlib import Path
from .doctsobjecttyperelation import DocTsObjectTypeRelation
import fetchx.ioutils as ioutils
from .doctemplates import Palette, use_template


class DocDimension:
    def __init__(self, dimension_dict: dict, tsobjecttype: 'DocTsObjectType'):
        """
        Represents a Dimension of a Time Series Object Type in the metamodel.
        :param dimension_dict: Dictionary containing dimension properties.
        :param tsobjecttype: The parent DocTsObjectType instance.
        """
        super().__init__()
        self.alias = dimension_dict.get("alias", None)  # type: dict | None
        self.alias_code = self.alias.get("code", None) if self.alias else None
        self.tsObjectType_code = dimension_dict.get("tsObjectTypeCode", None)  # type: str | None
        self.tsObjectType = tsobjecttype  # type: DocTsObjectType | None

class DocTsObjectType(dict):
    def __init__(self, *args, **kwargs):
        """
        Represents a Time Series Object Type in the metamodel.
        Inherits from dict to store object type properties.
        """
        super().__init__(*args, **kwargs)
        self.dimensions = None  # type: list[DocDimension] | None
        self.tsdefinitions = []  # type: list['DocTsDefinition']
        self.relationsToThisTsObjectType = []  # type: list[DocTsObjectTypeRelation]
        self.relationsFromThisTsObjectType = []  # type: list[DocTsObjectTypeRelation]
        self._html = None  # type: str | None

    @property
    def color(self) -> str:
        """
        Get the color associated with this Time Series Object Type.
        :return: Color as a hex string.
        """
        return Palette.get_color(self.get("code", ""))

    @staticmethod
    def load_from_directory(base_path: Path) -> dict[str, 'DocTsObjectType']:
        """
        Load all Time Series Object Types from the specified directory.
        :param base_path: Base path where the metamodel is located.
        :return: A dictionary mapping object type codes to DocTsObjectType instances.
        """
        result = {}
        tsobjecttypes_path = base_path / "metamodel/tsObjectType"
        # Load all .json files
        files, directories = ioutils.search_folder(str(tsobjecttypes_path.resolve()), list_of_extensions=[".json"])
        for file in files:
            value = ioutils.load_json(file)
            code = Path(file).stem
            result[code] = DocTsObjectType(value)
        return result
    
    def to_string(self, language: str = "en") -> str:
        """
        Convert the Time Series Object Type to a string representation.
        :param language: Language code for localization.
        :return: String representation of the Time Series Object Type.
        """
        if self._html is not None:
            return self._html
        result = ""
        # generate tsobjecttype for specific language, e.g., 'en'
        result += use_template('TSObjectType_Begin', {'TSObjectTypeCode': self.get("code", "--")}, language=language)
        # Generate footer content
        footer_content = ""
        footer_content += f'<strong>Value Provider Type: </strong>{self.get("valueProviderType", "--")}, '
        footer_content += f'<strong>Value Provider API: </strong>{self.get("valueProviderApi", "--")}'
        # generate block with border color from parent tsobjecttype
        result += use_template('TSObjectType_Block',{
            'border_color': self.color,
            'footer': footer_content,
            'TSObjectTypeCode':self.get('code','--'),
            'color_scheme': 'building'
            }, language=language
        ) # type: ignore
        result += use_template('TSObjectType_BlockSectionStart', {}, language=language)
        # generate block content
        result += use_template('TSObjectType_BlockContent',{
            'name': self.get('name', '--'),
            'short_name': self.get('shortName', '--'),
            'description': self.get('desc', '--'),
            'tsdefinitions_count': str(len(self.tsdefinitions))
            }, language=language
        )
        # add dimensions
        if self.dimensions is not None and len(self.dimensions) > 0:
            result += use_template('TSObjectType_DimensionsBegin',{
                'dimensions_count': str(len(self.dimensions))
                }, language=language
            )
            # add each dimension
            for dimension in self.dimensions:
                dimension_code = dimension.tsObjectType_code if dimension.tsObjectType_code is not None else "--"
                dimension_alias = dimension.alias_code if dimension.alias_code is not None else "--"
                dimension_link = f'#{dimension_code}'
                dimension_name = dimension_code
                if dimension.alias_code is not None:
                    dimension_name += f' (alias: {dimension_alias})'
                result += use_template('TSObjectType_Dimension',{
                    'dimension_link': dimension_link,
                    'dimension': dimension_name
                    }, language=language
                )
            result += use_template('TSObjectType_DimensionsEnd', {}, language=language)
        result += use_template('TSObjectType_BlockContentEnd', {}, language=language)
        if len(self.relationsToThisTsObjectType) > 0 or len(self.relationsFromThisTsObjectType) > 0:
            # add predecessors section
            result += use_template('TSObjectType_TransformationsToSectionStart',{
                'predecessors_count': str(len(self.relationsToThisTsObjectType))
                }, language=language
            )
            for relation in self.relationsToThisTsObjectType:
                transformation_code = relation.get("code","--") # type: ignore
                predecessor_link = f'#{transformation_code}'
                result += use_template('TSObjectType_TransformationsToStart',{
                    'predecessor_link': predecessor_link,
                    'transformation_code': transformation_code
                    }, language=language
                )
            result += use_template('TSObjectType_TransformationsToEnd', {}, language=language)
            # add successors section
            result += use_template('TSObjectType_TransformationsFromSectionStart',{
                'successors_count': str(len(self.relationsFromThisTsObjectType))
                }, language=language
            )
            for relation in self.relationsFromThisTsObjectType:
                transformation_code = relation.get("code","--") # type: ignore
                successor_link = f'#{transformation_code}'
                result += use_template('TSObjectType_TransformationsFromStart',{
                    'successor_link': successor_link,
                    'transformation_code': transformation_code
                    }, language=language
                )
            result += use_template('TSObjectType_TransformationsFromEnd', {}, language=language)
        result += use_template('TSObjectType_End', {}, language=language)
        self._html = result
        return result
        


from pathlib import Path
from .doctsobjecttyperelation import DocTsObjectTypeRelation
import fetchx.ioutils as ioutils
from .doctsbusinesstype import DocTsBusinessType
from .doctstimeunit import DocTsTimeUnit

class DocTsType(dict):
    def __init__(self, *args, **kwargs):
        """
        Represents a Time Series Type in the metamodel.
        Inherits from dict to store type properties.
        """
        super().__init__(*args, **kwargs)
        self.tsbusinesstype = None  # type: DocTsBusinessType | None
        self.tstimeunit = None  # type: DocTsTimeUnit | None

    @staticmethod
    def load_from_directory(base_path: Path) -> dict[str, 'DocTsType']:
        """
        Load all Time Series Types from the specified directory.
        :param base_path: Base path where the metamodel is located.
        :return: A dictionary mapping type codes to DocTsType instances.
        """
        result = {}
        tstype_path = base_path / "metamodel/tsType"
        # Load all .json files
        files, directories = ioutils.search_folder(str(tstype_path.resolve()), list_of_extensions=[".json"])
        for file in files:
            value = ioutils.load_json(file)
            code = Path(file).stem
            result[code] = DocTsType(value)
        return result
        


from pathlib import Path
import fetchx.ioutils as ioutils
import re


class DocTsFormula(dict):
    def __init__(self, *args, **kwargs):
        """
        Represents a Time Series Formula in the metamodel.
        Inherits from dict to store formula properties.
        """
        super().__init__(*args, **kwargs)
        self.tsdefinition = None

    @staticmethod
    def load_from_directory(tsformulas_path: Path) -> dict[str, 'DocTsFormula']:
        """
        Load all Time Series Formulas from the specified directory.
        :param tsformulas_path: Base path where the metamodel is located.
        :return: A dictionary mapping formula codes to DocTsFormula instances.
        """
        result = {}
        tsformulas_path = tsformulas_path / "calc/tsFormula"
        # Load all .json files
        files, directories = ioutils.search_folder(str(tsformulas_path.resolve()), list_of_extensions=[".json"])
        for file in files:
            value = ioutils.load_json(file)
            code = Path(file).stem
            result[code] = DocTsFormula(value)
        return result
    
    def parse_formula_words(self) -> list[str]:
        """
        Parse the formula and return a list of words (alphanumeric sequences and underscores).
        This is a simple parser that does not handle comments or functions.
        """
        formula = str(self.get("formula", ""))
        formula += " "  # add space at the end to capture last word
        words = []
        current_word = ""
        started_with_at = False
        for letter in formula:
            if letter.isalnum() or letter == '_':
                current_word += letter
            elif letter == '-' and started_with_at:
                current_word += letter
            elif letter == '@' and not current_word:
                current_word += letter
                started_with_at = True
            else:
                # add current word to list if exists
                if current_word:
                    if started_with_at:
                        words.append(current_word[1:])  # remove @ at the beginning
                    else:
                        words.append(current_word)
                # reset current word
                current_word = ""
                started_with_at = False
        return words
    
    def parse_formula_words2(self) -> list[str]:
        """
        Parse the formula and return a list of words (alphanumeric sequences and underscores).
        This is a simple parser that does not handle comments or functions.
        """
        # TODO process comments in formulas and ignore them
        formula = str(self.get("formula", ""))
        # Find all words (sequences of alphanumeric characters and underscores)
        words = re.findall(r'\b\w+\b', formula)
        # remove @ at the beginning of words if exists
        words = [word[1:] if word.startswith('@') else word for word in words]
        return words
        


from pathlib import Path
from .doctsobjecttyperelation import DocTsObjectTypeRelation
import fetchx.ioutils as ioutils


class DocTsTimeUnit(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        """
        Return string representation of the time unit
        """
        # Get the code and name of the time unit
        result = self.get('code','--')
        name = self.get('name',{})
        if isinstance(name, dict):
            name = name.get('en','--')
        else:
            name = '--'
        result += f' ({name})'
        return result

    @staticmethod
    def load_from_directory(base_path: Path) -> dict[str, 'DocTsTimeUnit']:
        """
        Load all Time Series Time Units from the specified directory.
        :param base_path: Base path where the metamodel is located.
        :return: A dictionary mapping time unit codes to DocTsTimeUnit instances.
        """
        result = {}
        tstype_path = base_path / "metamodel/tsTimeUnit"
        # Load all .json files
        files, directories = ioutils.search_folder(str(tstype_path.resolve()), list_of_extensions=[".json"])
        for file in files:
            value = ioutils.load_json(file)
            code = Path(file).stem
            result[code] = DocTsTimeUnit(value)
        return result
        


import os
import base64
from pathlib import Path
from .init_files import init_files

def _safe_copy_files():
    """
    Extracts and saves files from init_files.py to appropriate relative paths.
    Creates directory structure if it does not exist.
    Files are not overwritten if they already exist.
    """
    # Determine destination paths
    dest_path = Path.cwd().resolve()
    print(f'\nInitializing directory "{dest_path}"')
    
    # Process each file entry
    for relative_path, content_base64 in init_files.items():
        # Create the full destination path
        dest_file = dest_path / relative_path
        
        try:
            # Skip if file already exists
            if dest_file.exists():
                print(f'[SKIP]  File "{relative_path}" already exists, skipping copy.')
                continue
            
            # Create parent directories if they don't exist
            dest_file.parent.mkdir(parents=True, exist_ok=True)
            
            # Decode base64 content and write to file
            file_content = base64.b64decode(content_base64)
            with open(dest_file, "wb") as fdst:
                fdst.write(file_content)
            print(f'[OK]    file "{relative_path}" copied successfully.')
        except Exception as err:
            print(f'[ERROR] Could not create file "{dest_file}": {err}')   

def init():
    """
    Initializes the current working directory by copying necessary files from the
    init_files directory. Existing files are not overwritten.
    """
    try:
        _safe_copy_files()
        print('[DONE]  Initialization completed successfully.')
    except Exception as err:
        print(f'\n[ERROR] Initialization failed: {err}')
        raise


init()
